@@ -8,7 +8,7 @@ from dataclasses import dataclass
 from math import pi as PI
 import os
 from pathlib import Path
-from typing import Callable
+from typing import Callable, Optional
 from typing import Tuple
 import logging
 
@@ -75,6 +75,7 @@ class Dynamics(Model):
         super(Dynamics, self).__init__()
         # TODO: Implement reversibility check
         self.config = config
+        self.group = config.group
         self.xdim = self.config.xdim
         self.xshape = tuple(network_factory.input_spec.xshape)
         self.potential_fn = potential_fn
@@ -136,8 +137,9 @@ class Dynamics(Model):
     def apply_transition_hmc(
             self,
             inputs: tuple[Tensor, Tensor],
+            eps: Tensor,
     ) -> tuple[Tensor, dict]:
-        data = self.generate_proposal_hmc(inputs)
+        data = self.generate_proposal_hmc(inputs, eps)
         ma_, mr_ = self._get_accept_masks(data['metrics']['acc'])
         ma = ma_[:, None]
         mr = mr_[:, None]
@@ -236,11 +238,12 @@ class Dynamics(Model):
     def generate_proposal_hmc(
             self,
             inputs: tuple[Tensor, Tensor],
+            eps: Tensor,
     ) -> dict:
         x, beta = inputs
         v = tf.random.normal(x.shape, dtype=x.dtype)
         init = State(x, v, beta)
-        proposed, metrics = self.transition_kernel_hmc(init)
+        proposed, metrics = self.transition_kernel_hmc(init, eps=eps)
 
         return {'init': init, 'proposed': proposed, 'metrics': metrics}
 
@@ -287,7 +290,7 @@ class Dynamics(Model):
             self,
             state: State,
             logdet: Tensor,
-            step: int = None,
+            step: Optional[int] = None,
     ) -> dict:
         """Returns dict of various metrics about input State."""
         energy = self.hamiltonian(state)
@@ -322,22 +325,20 @@ class Dynamics(Model):
     def leapfrog_hmc(
             self,
             state: State,
-            step: int,
+            eps: Tensor,
     ) -> State:
         """Perform standard HMC leapfrog update."""
-        force1 = self.grad_potential(state.x, state.beta)
-        v1 = state.v - 0.5 * self.veps[step] * force1
-
-        xp = state.x + self.xeps[step] * v1
-
-        force2 = self.grad_potential(xp, state.beta)
-        v2 = v1 - 0.5 * self.veps[step] * force2
-
-        return State(x=xp, v=v2, beta=state.beta)
+        force1 = self.grad_potential(state.x, state.beta)  # f = dU / dx
+        v1 = state.v - 0.5 * tf.multiply(eps, force1)     # v -= ½ veps * f
+        xp = state.x + eps * v1                           # x += xeps * v
+        force2 = self.grad_potential(xp, state.beta)       # calc force, again
+        v2 = v1 - 0.5 * tf.multiply(eps, force2)          # v -= ½ veps * f
+        return State(x=xp, v=v2, beta=state.beta)          # output: (x', v')
 
     def transition_kernel_hmc(
             self,
             state: State,
+            eps: Tensor,
     ) -> tuple[State, dict]:
         """Run the generic HMC transition kernel."""
         state_ = State(x=state.x, v=state.v, beta=state.beta)
@@ -345,7 +346,7 @@ class Dynamics(Model):
         metrics = self.get_metrics(state_, sumlogdet)
         history = self.update_history(metrics, history={})
         for step in range(self.config.nleapfrog):
-            state_ = self.leapfrog_hmc(state_, step)
+            state_ = self.leapfrog_hmc(state_, eps=eps)
             if self.config.verbose:
                 metrics = self.get_metrics(state_, sumlogdet, step=step)
                 history = self.update_history(metrics, history=history)
