@@ -8,6 +8,7 @@ from dataclasses import asdict
 import logging
 import os
 from pathlib import Path
+import wandb
 import time
 from typing import Callable, Any, Optional
 
@@ -32,6 +33,8 @@ from l2hmc.utils.step_timer import StepTimer
 # from torchinfo import summary as model_summary
 
 
+WIDTH = int(os.environ.get('COLUMNS', 150))
+
 log = logging.getLogger(__name__)
 
 
@@ -83,10 +86,10 @@ class Trainer:
             schedule: AnnealingSchedule,
             lr_config: LearningRateConfig,
             loss_fn: Callable = LatticeLoss,
-            keep: str | list[str] = None,
-            skip: str | list[str] = None,
             aux_weight: float = 0.0,
-            dynamics_config: DynamicsConfig = None,
+            keep: Optional[str | list[str]] = None,
+            skip: Optional[str | list[str]] = None,
+            dynamics_config: Optional[DynamicsConfig] = None,
     ) -> None:
         self.steps = steps
         self.dynamics = dynamics
@@ -94,6 +97,7 @@ class Trainer:
         self.schedule = schedule
         self.loss_fn = loss_fn
         self.aux_weight = aux_weight
+        self.clip_norm = lr_config.clip_norm
         self._with_cuda = torch.cuda.is_available()
         self.accelerator = accelerator
         self.lr_config = lr_config
@@ -169,22 +173,27 @@ class Trainer:
 
         return m
 
-    def hmc_step(self, inputs: tuple[Tensor, float]) -> tuple[Tensor, dict]:
-        xinit, beta = inputs
-        xinit = to_u1(xinit).to(self.accelerator.device)
-        beta = torch.tensor(beta)
+    def hmc_step(
+            self,
+            inputs: tuple[Tensor, float],
+            eps: Tensor,
+    ) -> tuple[Tensor, dict]:
+        xi, beta = inputs
+        xi = to_u1(xi).to(self.accelerator.device)
+        beta = torch.tensor(beta).to(self.accelerator.device)
+        eps = eps.to(self.accelerator.device)
         # beta = torch.tensor(beta).to(self.accelerator.device)
-        xout, metrics = self._dynamics.apply_transition_hmc(  # type: ignore
-            (xinit, beta)
+        xo, metrics = self._dynamics.apply_transition_hmc(  # type: ignore
+            (xi, beta), eps=eps
         )
-        xprop = to_u1(metrics.pop('mc_states').proposed.x)
-        loss = self.loss_fn(x_init=xinit, x_prop=xprop, acc=metrics['acc'])
-        lmetrics = self.loss_fn.lattice_metrics(xinit=to_u1(xinit),
-                                                xout=to_u1(xout))
+        xo = to_u1(xo)
+        xp = to_u1(metrics.pop('mc_states').proposed.x)
+        loss = self.loss_fn(x_init=xi, x_prop=xp, acc=metrics['acc'])
+        lmetrics = self.loss_fn.lattice_metrics(xinit=xi, xout=xo)
         metrics.update(lmetrics)
         metrics.update({'loss': loss.detach().cpu().numpy()})
 
-        return to_u1(xout).detach(), metrics
+        return xo.detach(), metrics
 
     def eval_step(self, inputs: tuple[Tensor, float]) -> tuple[Tensor, dict]:
         xinit, beta = inputs
@@ -195,22 +204,20 @@ class Trainer:
         lmetrics = self.loss_fn.lattice_metrics(xinit=to_u1(xinit),
                                                 xout=to_u1(xout))
         metrics.update(lmetrics)
-        # metrics.update(lattice_metrics)
         metrics.update({'loss': loss.detach().cpu().numpy()})
 
         return to_u1(xout).detach(), metrics
 
     def eval(
             self,
-            beta: float = None,
-            x: Tensor = None,
-            skip: str | list[str] = None,
-            width: int = 150,
-            # eval_dir: os.PathLike = None,
+            beta: Optional[float] = None,
+            x: Optional[Tensor] = None,
+            skip: Optional[str | list[str]] = None,
             run: Optional[Any] = None,
             writer: Optional[Any] = None,
             job_type: Optional[str] = 'eval',
-            # hmc: bool = False,
+            nchains: Optional[int] = -1,
+            eps: Optional[Tensor] = None,
     ) -> dict:
         summaries = []
         self.dynamics.eval()
@@ -224,26 +231,47 @@ class Trainer:
             x = random_angle(self.xshape)
             x = x.reshape(x.shape[0], -1)
 
-        if job_type == 'hmc':
-            eval_fn = self.hmc_step
-        else:
-            eval_fn = self.eval_step
+        if eps is None and str(job_type).lower() == 'hmc':
+            eps = torch.tensor(0.1)
+            log.warn(
+                'Step size `eps` not specified for HMC! Using default: 0.1'
+            )
+
+        assert job_type in ['eval', 'hmc']
+
+        def eval_fn(z):
+            if job_type == 'hmc':
+                assert eps is not None
+                return self.hmc_step(z, eps)
+            return self.eval_step(z)
 
         summaries = []
         tables = {}
-        table = Table(row_styles=['dim', 'none'], box=box.SIMPLE)
-        # nlog = max(10, int(self.steps.test // 500))
-        # nprint = max(10, int(self.steps.test // 20))
-        nprint = self.steps.test // 20
-        nlog = 10 if self.steps.test < 1000 else 20
+        table = Table(row_styles=['dim', 'none'], box=box.HORIZONTALS)
+        nprint = max((20, self.steps.test // 20))
+        nlog = max((1, min((10, self.steps.test))))
+        if nlog <= self.steps.test:
+            nlog = min(10, max(1, self.steps.test // 100))
+
         assert job_type in ['eval', 'hmc']
         timer = self.timers[job_type]
         history = self.histories[job_type]
-        assert isinstance(beta, float)
 
-        with Live(table, console=console, screen=False) as live:
-            if width is not None and width > 0:
-                live.console.width = width
+        log.warning(f'x.shape (original): {x.shape}')
+        if nchains is not None:
+            if isinstance(nchains, int) and nchains > 0:
+                x = x[:nchains]
+
+        assert isinstance(x, Tensor)
+        assert isinstance(beta, float)
+        log.warning(f'x[:nchains].shape: {x.shape}')
+
+        if run is not None:
+            run.config.update({job_type: {'beta': beta, 'xshape': x.shape}})
+
+        with Live(table, screen=False, auto_refresh=False) as live:
+            if WIDTH is not None and WIDTH > 0:
+                live.console.width = WIDTH
 
             for step in range(self.steps.test):
                 timer.start()
@@ -251,29 +279,27 @@ class Trainer:
                 dt = timer.stop()
                 if step % nlog == 0 or step % nprint == 0:
                     record = {
-                        'step': step, 'dt': dt, 'loss': metrics['loss'],
-                        'dQint': metrics['dQint'], 'dQsin': metrics['dQsin'],
+                        'step': step, 'beta': beta, 'dt': dt,
                     }
-                    record.update(self.metrics_to_numpy(metrics))
-                    avgs = history.update(record)
-                    summary = summarize_dict(avgs)
+                    avgs, summary = self.record_metrics(run=run,
+                                                        step=step,
+                                                        record=record,
+                                                        writer=writer,
+                                                        metrics=metrics,
+                                                        history=history,
+                                                        job_type=job_type)
+                    if avgs.get('acc', 1.0) < 1e-5:
+                        log.warning('Chains are stuck! Re-drawing x !')
+                        x = random_angle(self.xshape)
+                        x = x.reshape(x.shape[0], -1)
+
                     summaries.append(summary)
-                    if writer is not None:
-                        update_summaries(step=step,
-                                         prefix=job_type,
-                                         metrics=record,
-                                         writer=writer)
-                        writer.flush()
-
-                    if run is not None:
-                        run.log({f'wandb/{job_type}': record}, commit=False)
-                        run.log({f'avgs/wandb.{job_type}': avgs})
-
                     if step == 0:
                         table = add_columns(avgs, table)
 
                     if step % nprint == 0:
                         table.add_row(*[f'{v:5}' for _, v in avgs.items()])
+                        live.refresh()
 
             tables[str(0)] = table
 
@@ -296,16 +322,88 @@ class Trainer:
             and self.accelerator.is_local_main_process
         )
 
-    def train_step(self, inputs: tuple[Tensor, float]) -> tuple[Tensor, dict]:
+    def record_metrics(
+            self,
+            metrics: dict,
+            job_type: str,
+            step: Optional[int] = None,
+            record: Optional[dict] = None,
+            run: Optional[Any] = None,
+            writer: Optional[Any] = None,
+            history: Optional[BaseHistory] = None,
+    ):
+        record = {} if record is None else record
+
+        if step is not None:
+            record.update({f'{job_type}_step': step})
+
+        record.update({
+            'loss': metrics.get('loss', None),
+            'dQint': metrics.get('dQint', None),
+            'dQsin': metrics.get('dQsin', None),
+        })
+
+        record.update(self.metrics_to_numpy(metrics))
+        if history is not None:
+            avgs = history.update(record)
+        else:
+            avgs = {k: v.mean() for k, v in record.items()}
+
+        summary = summarize_dict(avgs)
+        if writer is not None:
+            assert step is not None
+            update_summaries(step=step,
+                             prefix=job_type,
+                             metrics=record,
+                             writer=writer)
+            writer.flush()
+
+        if run is not None:
+            dQint = record.get('dQint', None)
+            if dQint is not None:
+                dQdict = {
+                    f'dQint/{job_type}': {
+                        'val': dQint,
+                        'step': step,
+                        'avg': dQint.mean(),
+                    }
+                }
+                run.log(dQdict, commit=False)
+
+            run.log({f'wandb/{job_type}': record}, commit=False)
+            run.log({f'avgs/wandb.{job_type}': avgs})
+
+        return avgs, summary
+
+    def profile_step(
+            self,
+            inputs: tuple[Tensor, Tensor]
+    ) -> tuple[Tensor, dict]:
         xinit, beta = inputs
         xinit = to_u1(xinit).to(self.accelerator.device)
+        beta = beta.to(self.accelerator.device)
         xout, metrics = self.dynamics((xinit, beta))
-        xout = to_u1(xout)
         xprop = to_u1(metrics.pop('mc_states').proposed.x)
         loss = self.loss_fn(x_init=xinit, x_prop=xprop, acc=metrics['acc'])
+        self.optimizer.zero_grad
+        self.accelerator.backward(loss)
+        self.accelerator.clip_grad_norm_(self.dynamics.parameters(),
+                                         max_norm=self.clip_norm,)
+        self.optimizer.step()
+
+        return to_u1(xout).detach(), metrics
+
+    def train_step(self, inputs: tuple[Tensor, Tensor]) -> tuple[Tensor, dict]:
+        xinit, beta = inputs
+        xinit = to_u1(xinit).to(self.accelerator.device)
+        beta = torch.tensor(beta).to(self.accelerator.device)
+        xout, metrics = self.dynamics((xinit, beta))
+        xprop = to_u1(metrics.pop('mc_states').proposed.x)
+        loss = self.loss_fn(x_init=xinit, x_prop=xprop, acc=metrics['acc'])
+        xout = to_u1(xout)
 
         if self.aux_weight > 0:
-            yinit = to_u1(self.draw_x())
+            yinit = to_u1(self.draw_x()).to(self.accelerator.device)
             _, metrics_ = self.dynamics((yinit, beta))
             yprop = to_u1(metrics_.pop('mc_states').proposed.x)
             aux_loss = self.aux_weight * self.loss_fn(x_init=yinit,
@@ -315,50 +413,88 @@ class Trainer:
 
         self.optimizer.zero_grad()
         self.accelerator.backward(loss)
+        # extract_model_from_parallel(self.dynamics).parameters(),
+        self.accelerator.clip_grad_norm_(
+            self.dynamics.parameters(),
+            max_norm=self.clip_norm,
+        )
         self.optimizer.step()
-        metrics.update(self.loss_fn.lattice_metrics(xinit=xinit, xout=xout))
-        # lmetrics = self.loss_fn.lattice_metrics(xinit=xinit, xout=xout)
-        # metrics.update(lmetrics)
-        metrics.update({'loss': loss.detach().cpu().numpy()})
+
+        metrics['loss'] = loss
+        lmetrics = self.loss_fn.lattice_metrics(xinit=xinit, xout=xout)
+        metrics.update(lmetrics)
 
         return xout.detach(), metrics
 
-    def save_ckpt(self, era, epoch, train_dir, **kwargs) -> None:
+    def save_ckpt(
+            self,
+            era: int,
+            epoch: int,
+            train_dir: os.PathLike,
+            metrics: Optional[dict] = None,
+            run: Optional[Any] = None,
+    ) -> None:
         dynamics = extract_model_from_parallel(self.dynamics)
         ckpt_dir = Path(train_dir).joinpath('checkpoints')
         ckpt_dir.mkdir(exist_ok=True, parents=True)
         ckpt_file = ckpt_dir.joinpath(f'ckpt-{era}-{epoch}.tar')
         log.info(f'Saving checkpoint to: {ckpt_file.as_posix()}')
-        dynamics.save(train_dir)  # type: ignore
+        # self.dynamics.save(train_dir)  # type: ignore
         xeps = {
-            k: grab(v) for k, v in dynamics.xeps.items()  # type:ignore
+            k: grab(v) for k, v in self.dynamics.xeps.items()  # type:ignore
         }
         veps = {
-            k: grab(v) for k, v in dynamics.veps.items()  # type:ignore
+            k: grab(v) for k, v in self.dynamics.veps.items()  # type:ignore
         }
-        torch.save({
+        ckpt = {
             'era': era,
             'epoch': epoch,
             'xeps': xeps,
             'veps': veps,
-            'model_state_dict': dynamics.state_dict(),  # type: ignore
+            'model_state_dict': self.dynamics.state_dict(),
             'optimizer_state_dict': self.optimizer.state_dict(),
-            **kwargs,
-        }, ckpt_file)
+        }
+        if metrics is not None:
+            ckpt.update(metrics)
+
+        # torch.save(ckpt, ckpt_file)
+        self.accelerator.save(ckpt, ckpt_file)
+        if run is not None:
+            assert run is wandb.run
+            outfile = Path(train_dir).joinpath('model.pth').as_posix()
+            # torch.save(self.dynamics.state_dict(), outfile.as_posix())
+            self.accelerator.save(self.dynamics.state_dict(), outfile)
+            artifact = wandb.Artifact('model', type='model')
+            artifact.add_file(outfile)
+            run.log_artifact(artifact)
+
+    def profile(self, nsteps: int = 5) -> dict:
+        self.dynamics.train()
+        x = self.draw_x()
+        beta = torch.tensor(1.0)
+        for _ in range(nsteps):
+            x, metrics = self.profile_step((x, beta))
+
+        return metrics
+
 
     def train(
             self,
-            x: Tensor = None,
-            skip: str | list[str] = None,
-            width: int = None,
-            train_dir: os.PathLike = None,
-            run: Any = None,
-            writer: Any = None,
+            x: Optional[Tensor] = None,
+            skip: Optional[str | list[str]] = None,
+            train_dir: Optional[os.PathLike] = None,
+            run: Optional[Any] = None,
+            writer: Optional[Any] = None,
             # keep: str | list[str] = None,
     ) -> dict:
         skip = [skip] if isinstance(skip, str) else skip
-        if width is None:
-            width = max((150, int(os.environ.get('COLUMNS', 150))))
+
+        if train_dir is None:
+            train_dir = Path(os.getcwd()).joinpath('train')
+        else:
+            train_dir = Path(train_dir)
+
+        train_dir.mkdir(exist_ok=True, parents=True)
 
         if isinstance(skip, str):
             skip = [skip]
@@ -372,60 +508,64 @@ class Trainer:
         epoch = 0
         tables = {}
         metrics = {}
+        rows = {}
         summaries = []
         timer = self.timers['train']
         history = self.histories['train']
         tkwargs = {
-            'box': box.SIMPLE,
+            'box': box.HORIZONTALS,
             'row_styles': ['dim', 'none'],
         }
         self.dynamics.train()
+        log.warning(f'x.dtype: {x.dtype}')
         for era in range(self.steps.nera):
             table = Table(**tkwargs)
             beta = self.schedule.betas[str(era)]
             if self.accelerator.is_local_main_process:
-                console.width = width
-                console.rule(f'ERA: {era}, BETA: {beta}')
+                console.width = WIDTH
+                console.print('\n')
+                console.rule(
+                    f'[red]ERA: {era} / {self.steps.nera}, BETA: {beta}[/]',
+                )
+                console.print('\n')
 
             with Live(
                     table,
                     screen=False,
-                    console=console,
+                    # console=console,
                     # refresh_per_second=1,
             ) as live:
+                if WIDTH is not None and WIDTH > 0:
+                    live.console.width = WIDTH
+
                 estart = time.time()
                 for epoch in range(self.steps.nepoch):
                     timer.start()
                     x, metrics = self.train_step((x, beta))
-                    gstep += 1
                     dt = timer.stop()
+                    gstep += 1
                     if self.should_print(epoch) or self.should_log(epoch):
                         record = {
-                            'era': era, 'epoch': epoch,
-                            'beta': beta, 'dt': dt,
-                            'loss': metrics['loss'],
-                            'dQint': metrics['dQint'],
-                            'dQsin': metrics['dQsin'],
+                            'era': era, 'epoch': epoch, 'beta': beta, 'dt': dt,
                         }
-                        # Update metrics with train step metrics, tmetrics
-                        record.update(self.metrics_to_numpy(metrics))
-                        if writer is not None:
-                            update_summaries(writer=writer,
-                                             # model=dynamics,  # type:ignore
-                                             step=gstep,
-                                             metrics=record,
-                                             prefix='train')
-                            writer.flush()
-
-                        avgs = history.update(record)
-                        summary = summarize_dict(avgs)
+                        avgs, summary = self.record_metrics(run=run,
+                                                            step=gstep,
+                                                            writer=writer,
+                                                            record=record,
+                                                            metrics=metrics,
+                                                            job_type='train',
+                                                            history=history)
+                        rows[gstep] = avgs
                         summaries.append(summary)
-                        if run is not None:
-                            run.log({'wandb/train': record}, commit=False)
-                            run.log({'avgs/wandb.train': avgs})
+
+                        if avgs.get('acc', 1.0) < 1e-5:
+                            log.warning('Chains are stuck! Re-drawing x !')
+                            x = random_angle(self.xshape)
+                            x = x.reshape(x.shape[0], -1)
 
                         if epoch == 0:
                             table = add_columns(avgs, table)
+
                         if self.should_print(epoch):
                             table.add_row(*[f'{v}' for _, v in avgs.items()])
 
@@ -434,17 +574,24 @@ class Trainer:
                 if writer is not None:
                     model = extract_model_from_parallel(self.dynamics)
                     model = model if isinstance(model, Module) else None
-                    update_summaries(writer=writer, step=gstep, model=model)
+                    update_summaries(step=gstep,
+                                     writer=writer,
+                                     model=self.dynamics)
 
-                self.save_ckpt(era, epoch, train_dir, loss=metrics['loss'])
-                live.console.log(
-                    f'Era {era} took: {time.time() - estart:<5g}s\n',
+                ckpt_metrics = {'loss': metrics['loss']}
+                self.save_ckpt(era, epoch, train_dir,
+                               metrics=ckpt_metrics, run=run)
+                log.info(
+                    f'Era {era} took: {time.time() - estart:<5g}s',
+                )
+                log.info(
                     f'Avgs over last era:\n {self.history.era_summary(era)}',
                 )
 
         return {
-            # 'xarr': xarr,
+            'timer': timer,
+            'rows': rows,
             'summaries': summaries,
-            'history': self.history,
+            'history': history,
             'tables': tables,
         }
