@@ -5,9 +5,9 @@ Implements methods for training L2HMC sampler
 """
 from __future__ import absolute_import, annotations, division, print_function
 import logging
-from math import pi as PI
 import os
 from pathlib import Path
+# import wandb
 import time
 from typing import Callable, Any, Optional
 
@@ -33,9 +33,8 @@ from l2hmc.trackers.tensorflow.trackers import update_summaries
 from l2hmc.utils.step_timer import StepTimer
 from l2hmc.utils.tensorflow.history import History
 
-# import wandb
-# from rich.panel import Panel
-# from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn
+
+WIDTH = int(os.environ.get('COLUMNS', 150))
 
 tf.autograph.set_verbosity(0)
 os.environ['AUTOGRAPH_VERBOSITY'] = '0'
@@ -112,6 +111,27 @@ def plot_models(dynamics: Dynamics, logdir: os.PathLike):
             pass
 
 
+HVD_FP_MAP = {
+    'fp16': hvd.Compression.fp16,
+    'none': hvd.Compression.none,
+}
+
+
+def reset_optimizer(optimizer: Optimizer):
+    """Reset optimizer states when changing beta during training."""
+    # --------------------------------------------------------------
+    # NOTE: We don't want to reset iteration counter. From tf docs:
+    # > The first value is always the iterations count of the optimizer,
+    # > followed by the optimizer's state variables in the order they are
+    # > created.
+    # --------------------------------------------------------------
+    weight_shapes = [x.shape for x in optimizer.get_weights()[1:]]
+    optimizer.set_weights([
+        tf.zeros_like(x) for x in weight_shapes
+    ])
+    return optimizer
+
+
 class Trainer:
     def __init__(
             self,
@@ -123,26 +143,25 @@ class Trainer:
             rank: int = 0,
             loss_fn: Callable = LatticeLoss,
             aux_weight: float = 0.0,
-            keep: str | list[str] = None,
-            skip: str | list[str] = None,
-            compression: bool = True,
+            keep: Optional[str | list[str]] = None,
+            skip: Optional[str | list[str]] = None,
+            compression: Optional[str] = 'none',
             evals_per_step: int = 1,
-            dynamics_config: DynamicsConfig = None,
+            dynamics_config: Optional[DynamicsConfig] = None,
     ) -> None:
         self.rank = rank
         self.steps = steps
-        self.dynamics = dynamics
-        self.optimizer = optimizer
         self.loss_fn = loss_fn
+        self.dynamics = dynamics
         self.schedule = schedule
+        self.lr_config = lr_config
+        self.optimizer = optimizer
         self.aux_weight = aux_weight
+        self.clip_norm = lr_config.clip_norm
         self.keep = [keep] if isinstance(keep, str) else keep
         self.skip = [skip] if isinstance(skip, str) else skip
-        if compression:
-            self.compression = hvd.Compression.fp16
-        else:
-            self.compression = hvd.Compression.none
-
+        assert compression in ['none', 'fp16']
+        self.compression = HVD_FP_MAP.get(compression, hvd.Compression.none)
         self.reduce_lr = ReduceLROnPlateau(lr_config)
         self.dynamics.compile(optimizer=self.optimizer, loss=self.loss_fn)
         self.reduce_lr.set_model(self.dynamics)
@@ -154,7 +173,6 @@ class Trainer:
         self.history = History(steps=steps)
         evals_per_step = self.dynamics.config.nleapfrog
         self.timer = StepTimer(evals_per_step=evals_per_step)
-        self.eval_timer = StepTimer(evals_per_step=evals_per_step)
         self.histories = {
             'train': self.history,
             'eval': History(),
@@ -165,8 +183,14 @@ class Trainer:
             'eval': StepTimer(evals_per_step=evals_per_step),
             'hmc': StepTimer(evals_per_step=evals_per_step),
         }
-        # self.eval_history = History()
-        # evals_per_step = self.dynamics.config.nleapfrog * steps.log
+
+    def draw_x(self) -> Tensor:
+        """Draw `x` """
+        x = tf.random.uniform(self.dynamics.xshape,
+                              *(-4, 4), dtype=TF_FLOAT)
+        x = tf.reshape(x, (x.shape[0], -1))
+
+        return to_u1(x)
 
     def setup_CheckpointManager(self, outdir: os.PathLike):
         ckptdir = Path(outdir).joinpath('checkpoints')
@@ -191,15 +215,6 @@ class Trainer:
 
         return manager
 
-    def draw_x(self) -> Tensor:
-        """Draw `x` """
-        x = tf.random.uniform(self.dynamics.xshape,
-                              *(-PI, PI), dtype=TF_FLOAT)
-        x = tf.reshape(x, (x.shape[0], -1))
-
-        return x
-
-    # type: ignore
     def metric_to_numpy(
             self,
             metric: Tensor | list | np.ndarray,
@@ -269,11 +284,63 @@ class Trainer:
             and self.rank == 0
         )
 
+    def record_metrics(
+            self,
+            metrics: dict,
+            job_type: str,
+            step: Optional[int] = None,
+            record: Optional[dict] = None,
+            run: Optional[Any] = None,
+            writer: Optional[Any] = None,
+            history: Optional[History] = None,
+    ):
+        record = {} if record is None else record
+        if step is not None:
+            record.update({f'{job_type}_step': step})
+
+        record.update({
+            'loss': metrics.get('loss', None),
+            'dQint': metrics.get('dQint', None),
+            'dQsin': metrics.get('dQsin', None),
+        })
+        record.update(self.metrics_to_numpy(metrics))
+        if history is not None:
+            avgs = history.update(record)
+        else:
+            avgs = {k: v.mean() for k, v in record.items()}
+
+        summary = summarize_dict(avgs)
+        if writer is not None:
+            assert step is not None
+            update_summaries(step=step, prefix=job_type, metrics=record)
+            writer.flush()
+
+        if run is not None:
+            dQint = record.get('dQint', None)
+            if dQint is not None:
+                dQdict = {
+                    f'dQint/{job_type}': {
+                        'val': dQint,
+                        'step': step,
+                        'avg': dQint.mean()
+                    }
+                }
+                run.log(dQdict, commit=False)
+
+            run.log({f'wandb/{job_type}': record}, commit=False)
+            run.log({f'avgs/wandb.{job_type}': avgs})
+
+        return avgs, summary
+
     @tf.function(experimental_follow_type_hints=True)
-    def hmc_step(self, inputs: tuple[Tensor, Tensor]) -> tuple[Tensor, dict]:
+    def hmc_step(
+            self,
+            inputs: tuple[Tensor, Tensor],
+            eps: Tensor,
+    ) -> tuple[Tensor, dict]:
         xi, beta = inputs
-        xi = to_u1(xi)
-        xo, metrics = self.dynamics.apply_transition_hmc((xi, beta))
+        inputs = (to_u1(xi), tf.constant(beta))
+        xo, metrics = self.dynamics.apply_transition_hmc(inputs, eps=eps)
         xo = to_u1(xo)
         xp = to_u1(metrics.pop('mc_states').proposed.x)
         loss = self.loss_fn(x_init=xi, x_prop=xp, acc=metrics['acc'])
@@ -284,28 +351,32 @@ class Trainer:
         return xo, metrics
 
     @tf.function(experimental_follow_type_hints=True)
-    def eval_step(self, inputs: tuple[Tensor, Tensor]) -> tuple[Tensor, dict]:
-        xinit, beta = inputs
-        xout, metrics = self.dynamics((to_u1(xinit), tf.constant(beta)),
-                                      training=False)
-        xprop = to_u1(metrics.pop('mc_states').proposed.x)
-        loss = self.loss_fn(x_init=xinit, x_prop=xprop, acc=metrics['acc'])
-        lmetrics = self.loss_fn.lattice_metrics(xinit=to_u1(xinit),
-                                                xout=to_u1(xout))
+    def eval_step(
+            self,
+            inputs: tuple[Tensor, Tensor],
+    ) -> tuple[Tensor, dict]:
+        xi, beta = inputs
+        inputs = (to_u1(xi), tf.constant(beta))
+        xo, metrics = self.dynamics(inputs, training=False)
+        xo = to_u1(xo)
+        xp = to_u1(metrics.pop('mc_states').proposed.x)
+        loss = self.loss_fn(x_init=xi, x_prop=xp, acc=metrics['acc'])
+        lmetrics = self.loss_fn.lattice_metrics(xinit=xi, xout=xo)
         metrics.update(lmetrics)
         metrics.update({'loss': loss})
 
-        return to_u1(xout), metrics
+        return xo, metrics
 
     def eval(
             self,
-            beta: float = None,
-            x: Tensor = None,
-            skip: str | list[str] = None,
-            width: int = 150,
-            run: Any = None,
-            writer: Any = None,
+            beta: Optional[float] = None,
+            x: Optional[Tensor] = None,
+            skip: Optional[str | list[str]] = None,
+            run: Optional[Any] = None,
+            writer: Optional[Any] = None,
             job_type: Optional[str] = 'eval',
+            nchains: Optional[int] = None,
+            eps: Optional[Tensor] = None,
     ) -> dict:
         """Evaluate model."""
         if isinstance(skip, str):
@@ -314,11 +385,17 @@ class Trainer:
         if beta is None:
             beta = self.schedule.beta_final
 
+        if eps is None and str(job_type).lower() == 'hmc':
+            eps = tf.constant(0.1)
+            log.warn(
+                'Step size `eps` not specified for HMC! Using default: 0.1'
+            )
+
         assert job_type in ['eval', 'hmc']
 
         if x is None:
-            unif = tf.random.uniform(self.dynamics.xshape,
-                                     *(-PI, PI), dtype=TF_FLOAT)
+            unif = to_u1(tf.random.uniform(self.dynamics.xshape,
+                                           *(-4, 4), dtype=TF_FLOAT))
             x = tf.reshape(unif, (unif.shape[0], -1))
         else:
             x = tf.Variable(x, dtype=TF_FLOAT)
@@ -326,30 +403,42 @@ class Trainer:
         if writer is not None:
             writer.set_as_default()
 
-        if job_type == 'hmc':
-            eval_fn = self.hmc_step
-        else:
-            eval_fn = self.eval_step
+        def eval_fn(z):
+            if job_type == 'hmc':
+                return self.hmc_step(z, eps=eps)  # type: ignore
+            return self.eval_step(z)              # type: ignore
 
         assert isinstance(x, Tensor) and x.dtype == TF_FLOAT
 
         tables = {}
+        rows = {}
         summaries = []
-        table = Table(row_styles=['dim', 'none'], box=box.SIMPLE)
-        nprint = self.steps.test // 20
-        nlog = 10 if self.steps.test < 1000 else 20
+        table = Table(row_styles=['dim', 'none'], box=box.HORIZONTALS)
+        nprint = max((20, self.steps.test // 20))
+        nlog = max((1, min((10, self.steps.test))))
+        if nlog <= self.steps.test:
+            nlog = min(10, max(1, self.steps.test // 100))
+
         assert job_type in ['eval', 'hmc']
         timer = self.timers[job_type]
         history = self.histories[job_type]
-        with Live(
-            table,
-            console=console,
-            screen=False,
-            # refresh_per_second=4,
-            # auto_refresh=False,
-        ) as live:
-            if width is not None and width > 0:
-                live.console.width = width
+
+        log.warning(f'x.shape (original): {x.shape}')
+        if nchains is not None:
+            if isinstance(nchains, int) and nchains > 0:
+                x = x[:nchains]  # type: ignore
+
+        assert isinstance(x, Tensor)
+        log.warning(f'x[:nchains].shape: {x.shape}')
+
+        if run is not None:
+            run.config.update({
+                job_type: {'beta': beta, 'xshape': x.shape.as_list()}
+            })
+
+        with Live(table, screen=False, auto_refresh=False) as live:
+            if WIDTH is not None and WIDTH > 0:
+                live.console.width = WIDTH
 
             for step in range(self.steps.test):
                 timer.start()
@@ -357,34 +446,34 @@ class Trainer:
                 dt = timer.stop()
                 if step % nprint == 0 or step % nlog == 0:
                     record = {
-                        'step': step, 'dt': dt, 'loss': metrics['loss'],
-                        'dQint': metrics['dQint'], 'dQsin': metrics['dQsin'],
+                        'step': step, 'beta': beta, 'dt': dt,
                     }
-                    record.update(self.metrics_to_numpy(metrics))
-                    avgs = history.update(record)
-                    summary = summarize_dict(avgs)
+                    avgs, summary = self.record_metrics(run=run,
+                                                        step=step,
+                                                        record=record,
+                                                        writer=writer,
+                                                        metrics=metrics,
+                                                        history=history,
+                                                        job_type=job_type)
+                    rows[step] = avgs
                     summaries.append(summary)
-                    if writer is not None:
-                        update_summaries(step=step,
-                                         prefix=job_type,
-                                         metrics=record)
-                        writer.flush()
-
-                    if run is not None:
-                        run.log({f'wandb/{job_type}': record}, commit=False)
-                        run.log({f'avgs/wandb.{job_type}': avgs})
+                    if avgs.get('acc', 1.0) <= 1e-5:
+                        log.warning('Chains are stuck! Re-drawing x !')
+                        x = self.draw_x()
 
                     if step == 0:
                         table = add_columns(avgs, table)
 
                     if step % nprint == 0:
                         table.add_row(*[f'{v:5}' for _, v in avgs.items()])
+                        live.refresh()
 
             tables[str(0)] = table
 
         return {
             'timer': timer,
             'history': history,
+            'rows': rows,
             'summaries': summaries,
             'tables': tables,
         }
@@ -397,13 +486,12 @@ class Trainer:
     ) -> tuple[Tensor, dict]:
         xinit, beta = inputs
         xinit = to_u1(xinit)
-        # beta = tf.constant(beta)
-        record = {'loss': None}
         with tf.GradientTape() as tape:
             tape.watch(xinit)
             xout, metrics = self.dynamics((xinit, beta), training=True)
             xprop = to_u1(metrics.pop('mc_states').proposed.x)
             loss = self.loss_fn(x_init=xinit, x_prop=xprop, acc=metrics['acc'])
+            xout = to_u1(xout)
 
             if self.aux_weight > 0:
                 yinit = to_u1(self.draw_x())
@@ -414,54 +502,53 @@ class Trainer:
                                                           acc=metrics_['acc'])
                 loss = (loss + aux_loss) / (1. + self.aux_weight)
 
-            record['loss'] = loss
-            record.update(metrics)
-
         tape = hvd.DistributedGradientTape(tape, compression=self.compression)
-        grads = tape.gradient(loss, self.dynamics.trainable_variables)
-        updates = zip(grads, self.dynamics.trainable_variables)
-        self.optimizer.apply_gradients(updates)
+        grads = [
+            tf.clip_by_norm(grad, clip_norm=self.clip_norm)
+            for grad in tape.gradient(loss, self.dynamics.trainable_variables)
+        ]
+        self.optimizer.apply_gradients(
+            zip(grads, self.dynamics.trainable_variables)
+        )
         if first_step:
             hvd.broadcast_variables(self.dynamics.variables, root_rank=0)
             hvd.broadcast_variables(self.optimizer.variables(), root_rank=0)
 
-        # bconst = tf.cast(beta, dtype=TF_FLOAT)
-        xout = to_u1(xout)
+        metrics['loss'] = loss
         lmetrics = self.loss_fn.lattice_metrics(xinit=xinit, xout=xout)
-        record.update(lmetrics)
+        metrics.update(lmetrics)
 
-        return to_u1(xout), record
+        return to_u1(xout), metrics
 
     def train(
             self,
-            xinit: Tensor = None,
-            skip: str | list[str] = None,
-            width: int = None,
-            train_dir: os.PathLike = None,
-            run: Any = None,
-            writer: Any = None,
+            xinit: Optional[Tensor] = None,
+            skip: Optional[str | list[str]] = None,
+            train_dir: Optional[os.PathLike] = None,
+            run: Optional[Any] = None,
+            writer: Optional[Any] = None,
             # compile: bool = True,
             # jit_compile: bool = False,
             # save_x: bool = False,
     ) -> dict:
         """Train l2hmc Dynamics."""
-        if width is None:
-            width = max((150, int(os.environ.get('COLUMNS', 150))))
-
         if isinstance(skip, str):
             skip = [skip]
 
+        if writer is not None:
+            writer.set_as_default()
+
         if train_dir is None:
             train_dir = Path(os.getcwd()).joinpath('train')
 
-        x = self.draw_x() if xinit is None else tf.constant(xinit, TF_FLOAT)
-        inputs = (x, tf.constant(1.))
-        assert isinstance(x, Tensor) and x.dtype == TF_FLOAT
         manager = self.setup_CheckpointManager(train_dir)
-        # self.dynamics.compile(optimizer=self.optimizer, loss=self.loss_fn)
+        gstep = K.get_value(self.optimizer.iterations)
+
+        x = self.draw_x() if xinit is None else tf.constant(xinit, TF_FLOAT)
+        assert isinstance(x, Tensor) and x.dtype == TF_FLOAT
+
+        inputs = (x, tf.constant(1.))
         _ = self.train_step(inputs, first_step=True)  # type: ignore
-        if writer is not None:
-            writer.set_as_default()
 
         era = 0
         epoch = 0
@@ -472,22 +559,28 @@ class Trainer:
         timer = self.timers['train']
         history = self.histories['train']
         tkwargs = {
-            'box': box.SIMPLE,
+            'box': box.HORIZONTALS,
             'row_styles': ['dim', 'none'],
         }
-        gstep = K.get_value(self.optimizer.iterations)
         for era in range(self.steps.nera):
             table = Table(**tkwargs)
             beta = tf.constant(self.schedule.betas[str(era)])
             if self.rank == 0:
-                console.width = width
-                console.rule(f'ERA: {era}, BETA: {beta.numpy()}')
+                console.width = WIDTH
+                console.print('\n')
+                console.rule(
+                    f'[red]ERA: {era} / {self.steps.nera}, BETA: {beta}[/]',
+                )
+                console.print('\n')
+
             with Live(
                 table,
                 screen=False,
                 console=console,
                 # refresh_per_second=1,
             ) as live:
+                if WIDTH is not None and WIDTH > 0:
+                    live.console.width = WIDTH
                 estart = time.time()
                 for epoch in range(self.steps.nepoch):
                     timer.start()
@@ -499,29 +592,25 @@ class Trainer:
                     gstep += 1
                     if self.should_print(epoch) or self.should_log(epoch):
                         record = {
-                            'era': era, 'epoch': epoch,
-                            'beta': beta, 'dt': dt,
-                            'loss': metrics['loss'],
-                            'dQint': metrics['dQint'],
-                            'dQsin': metrics['dQsin'],
+                            'era': era, 'epoch': epoch, 'beta': beta, 'dt': dt,
                         }
-                        record.update(self.metrics_to_numpy(metrics))
-                        if writer is not None:
-                            update_summaries(step=gstep,
-                                             prefix='train',
-                                             metrics=record)
-                            writer.flush()
-
-                        avgs = history.update(record)
+                        avgs, summary = self.record_metrics(run=run,
+                                                            step=gstep,
+                                                            writer=writer,
+                                                            record=record,
+                                                            metrics=metrics,
+                                                            job_type='train',
+                                                            history=history)
                         rows[gstep] = avgs
-                        summary = summarize_dict(avgs)
                         summaries.append(summary)
-                        if run is not None:
-                            run.log({'wandb/train': record}, commit=False)
-                            run.log({'avgs/wandb.train': avgs})
+
+                        if avgs.get('acc', 1.0) <= 1e-5:
+                            log.warning('Chains are stuck! Re-drawing x !')
+                            x = self.draw_x()
 
                         if epoch == 0:
                             table = add_columns(avgs, table)
+
                         if self.should_print(epoch):
                             table.add_row(*[f'{v}' for _, v in avgs.items()])
 
@@ -534,13 +623,18 @@ class Trainer:
                     update_summaries(step=gstep,
                                      model=self.dynamics,
                                      optimizer=self.optimizer)
-                live.console.log(
-                    f'Era {era} took: {time.time() - estart:<5g}s\n',
+                log.info(
+                    f'Era {era} took: {time.time() - estart:<5g}s'
+                )
+                log.info(
                     f'Avgs over last era:\n {self.history.era_summary(era)}',
+                )
+                log.info(
                     f'Saving checkpoint to: {manager.latest_checkpoint}'
                 )
                 manager.save()
-                self.dynamics.save_networks(train_dir)
+                if (era + 1) == self.steps.nera or (era + 1) % 5 == 0:
+                    self.dynamics.save_networks(train_dir)
 
         return {
             'timer': timer,
